;**************************************************************
;     Parser
;
; Author: Josef Sniatecki
; Creation date: 03.04.2009
;**************************************************************
;   (c) 2009 Josef Sniatecki
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; As published by the Free Software Foundation; either version
; 3 of the License, Or (at your option) any later version.
;
; This program is distributed in the hope that it will be
; useful, but WITHOUT ANY WARRANTY; without even the implied
; warranty of MERCHANTABILITY Or FITNESS For A PARTICULAR
; PURPOSE. See the GNU General Public License For more details.
;
; You should have received a copy of the GNU General Public
; License along With this program;
; If Not, see "http://www.gnu.org/licenses/".
;**************************************************************
; This source code includes the parser and it's lexer. The
; lexer fragments a code into tokens. While this, the parser
; looks after preprocessor commands.



Procedure.i Parse(*TokenQueue.Queue, *Pointer.Character, SingleLine.i = #False)
  Protected Integer.i, String.s, *Start.Character, *Line.Character
  Protected *TokenQueueStack.Queue, *OtherTokenQueue.Queue, *SubTokenQueue.Queue
  Protected *ArgumentQueue.Queue, *Item.QueueItem
  
  *TokenQueueStack = CreateQueue(#Queue_SaveMemory)
  *Line            = *Pointer
  
  Repeat
    Select *Pointer\C
      Case 'a' To 'z', 'A' To 'Z', '_', 192 To 255, ''
        ; Read out name:
        *Start = *Pointer
        *Pointer + _Sz(Character)
        Repeat
          Select *Pointer\C
            Case 'a' To 'z', 'A' To 'Z', '_', '0' To '9', 192 To 255, '!', '$', '?', ''
              *Pointer + _Sz(Character)
            Default
              *Pointer - _Sz(Character)
              Break
          EndSelect
        ForEver
        String = PeekS(*Start, *Pointer - *Start + _Sz(Character))
        ; Check predefined keywords:
        Integer = KeywordID(String)
        If Integer <> -1
          ; Compare two keywords:
          With *TokenQueue
            If Integer = #Keyword_If And \Last And \Last\Token\Type = #Token_Keyword And \Last\Token\ID = #Keyword_Else
              ; "else" and "if" will be pieced to "elseIf":
              \Last\Token\ID = #Keyword_ElseIf
            Else
              ; Single keyword:
              AddIntegerToken(*TokenQueue, #Token_Keyword, Integer)
            EndIf
          EndWith
        Else
          ; Check predefined names:
          Select String
            Case "as"
              AddIntegerToken(*TokenQueue, #Token_Control, #Control_Type)
              *Pointer = Parser_IgnoreNewLine(*Pointer, @*Line)
            Case "begin"
              AddBlockToken(*TokenQueue, #Token_Block, #Bracket_Curly, 0)
              AppendQueueItem(*TokenQueueStack, *TokenQueue)
              *TokenQueue = CreateQueue(@FreeToken())
              *Pointer = Parser_IgnoreNewLine(*Pointer, @*Line)
            Case "finish"
              ; Close the current block:
              If *TokenQueueStack\Count
                With *TokenQueue
                  If \Last And \Last\Token\Type = #Token_Seperator And \Last\Token\ID = #Seperator_Command
                    \Last\Token\ID = #Seperator_Block
                  Else
                    AddIntegerToken(*TokenQueue, #Token_Seperator, #Seperator_Block)
                  EndIf
                  Integer = *TokenQueue
                  *TokenQueue = PopQueueItem(*TokenQueueStack)
                  \Last\Token\Block\TokenQueue = Integer
                  If \Last\Token\Block\Bracket <> #Bracket_Curly
                    _ErrorB("Two different bracktes used for the same block: " + #Q + Bracket(\Last\Token\Block\Bracket) + "...}" + #Q)
                  EndIf
                EndWith
              Else
                _ErrorB("Closing bracket without an opened block: " + #Q + "..." + Chr(*Pointer\C) + #Q)
              EndIf
            Case "method"
              Integer = 0
              *Start = 0
              *Pointer + _Sz(Character)
              Repeat
                Select *Pointer\C
                  Case ' '
                    ;Ignore
                  Case 'a' To 'z', 'A' To 'Z', '_', 192 To 255, ''
                    Integer = 1
                    ; Read out class method:
                    *Start = *Pointer
                    *Pointer + _Sz(Character)
                    Repeat
                      Select *Pointer\C
                        Case 'a' To 'z', 'A' To 'Z', '_', '0' To '9', 192 To 255, '!', '$', '?', ''
                          *Pointer + _Sz(Character)
                        Default
                          *Pointer - _Sz(Character)
                          Break
                      EndSelect
                    ForEver
                    String = PeekS(*Start, *Pointer - *Start + _Sz(Character))
                    If String = "friend"
                      ; Get friend method of an operation:
                      Integer = 3
                    Else
                      Break
                    EndIf
                  Case '+', '-', '*', '/', '=', '<', '>', '&', '%', '|'
                    If Not Integer
                      Integer = 2
                    EndIf
                    *Start = *Pointer
                    *Pointer + _Sz(Character)
                    Repeat
                      Select *Pointer\C
                        Case '+', '-', '*', '/', '=', '<', '>', '&', '%', '|', '!'
                          *Pointer + _Sz(Character)
                        Default
                          *Pointer - _Sz(Character)
                          Break
                      EndSelect
                    ForEver
                    String = PeekS(*Start, *Pointer - *Start + _Sz(Character))
                    Break
                EndSelect
                *Pointer + _Sz(Character)
              ForEver
              Select Integer
                Case 1
                  String = "_" + String
                  Integer = ClassMethodID(String)
                  If Integer = #None
                    _ErrorB("Class method not available: " + #Q + String + #Q)
                  EndIf
                  AddStringToken(*TokenQueue, #Token_Name, String)
                Case 2
                  Integer = OperatorID(String)
                  If Integer = #None
                    _ErrorB("Invalid operator: " + #Q + String + #Q)
                  EndIf
                  Integer = OperatorMethod(Integer)
                  If Integer = #None
                    _ErrorB("No class method available for this operator: " + #Q + String + #Q)
                  EndIf
                  AddStringToken(*TokenQueue, #Token_Name, ClassMethod(Integer))
                Case 3
                  Integer = OperatorID(String)
                  If Integer = #None
                    _ErrorB("Invalid operator: " + #Q + String + #Q)
                  EndIf
                  Integer = OperatorFriendMethod(Integer)
                  If Integer = #None
                    _ErrorB("No friend method available for this operator: " + #Q + String + #Q)
                  EndIf
                  AddStringToken(*TokenQueue, #Token_Name, ClassMethod(Integer))
              EndSelect
            Case "global"
              ; Become to "main\":
              AddStringToken(*TokenQueue, #Token_Name, "main")
              AddIntegerToken(*TokenQueue, #Token_Control, #Control_Next)
            Case "addressOf"
              AddIntegerToken(*TokenQueue, #Token_Control, #Control_Address)
            Default
              ; Self defined name:
              With *TokenQueue
                If \Last And \Last\Token\Type = #Token_Name
                  ; Compare both names to one name:
                  String = \Last\Token\Name + UCase(Left(String, 1)) + Mid(String, 2)
                  ; Check predefined compareable names:
                  Select String
                    Case "addressOf"
                      FreeLastQueueItem(*TokenQueue)
                      AddIntegerToken(*TokenQueue, #Token_Control, #Control_Address)
                    Default
                      \Last\Token\String = String
                  EndSelect
                Else
                  AddStringToken(*TokenQueue, #Token_Name, String)
                EndIf
              EndWith
          EndSelect
        EndIf
      Case '0' To '9'
        ; Number:
        String = Chr(*Pointer\C)
        Integer = 0 ;0 = Integer/Quad. 1 = Float/Double
        *Pointer + 1
        Select *Pointer\C
          Case '$'
            ; Binary:
            *Pointer + 1
            String = "%" ;Signals a binary number for "val"
            Repeat
              Select *Pointer\C
                Case '0', 'o', 'O', 'Q'
                  String + "0"
                Case '1', 'i', 'I'
                  String + "1"
                Default
                  *Pointer - 1
                  Break
              EndSelect
              *Pointer + 1
            ForEver
          Case '#'
            ; Hex:
            *Pointer + 1
            String = "$" ;Signals a hexdecimal number for "val"
            Repeat
              Select *Pointer\C
                Case '0' To '9', 'a' To 'f', 'A' To 'F'
                  String + Chr(*Pointer\C)
                Default
                  *Pointer - 1
                  Break
              EndSelect
              *Pointer + 1
            ForEver
          Case '0' To '9', '.', ''
            ; Decimal:
            Repeat
              Select *Pointer\C
                Case '0' To '9'
                  String + Chr(*Pointer\C)
                Case '.'
                  *Pointer + 1
                  If *Pointer\C <> '.'
                    *Pointer - 1
                    Integer = 1 ;Note, that the number is a float/double
                    String + "."
                  Else
                    *Pointer - 2
                    Break
                  EndIf
                Case ''
                  String = StrD(_Rad(ValD(String)))
                  Integer = 1
                  Break
                Default
                  *Pointer - 1
                  Break
              EndSelect
              *Pointer + 1
            ForEver
          Default
            ; Only null:
            *Pointer - 1
        EndSelect
        ; Notice a minus behind this number:
        With *TokenQueue
          If \Last And \Last\Token\Type = #Token_Operator And \Last\Token\Operator = #Operator_Subtract
            ; Check previous token:
            If \Last\P
              Select \Last\P\Token\Type
                Case #Token_Operator, #Token_Seperator, #Token_Keyword
                  ; Convert to a negative number:
                  FreeLastQueueItem(*TokenQueue)
                  String = "-" + String
              EndSelect
            Else
              ; Convert to a negative number:
              FreeLastQueueItem(*TokenQueue)
              String = "-" + String
            EndIf
          EndIf
        EndWith
        Select Integer
          Case 0
            ;Integer/Quad:
            If Len(String) < 8
              AddValueToken(*TokenQueue, #Token_Value, CreateIValue(Val(String)))
            Else
              AddValueToken(*TokenQueue, #Token_Value, CreateQValue(Val(String)))
            EndIf
          Case 1
            ;Float/Double:
            If Len(String) < 8
              AddValueToken(*TokenQueue, #Token_Value, CreateFValue(ValF(String)))
            Else
              AddValueToken(*TokenQueue, #Token_Value, CreateDValue(ValD(String)))
            EndIf
        EndSelect
      Case '"'
        ; String:
        *Pointer + 1
        String = ""
        *OtherTokenQueue = #Null
        
        If *Pointer\C = '"'
          *Pointer + 1
          If *Pointer\C = '"'
            ; Top-level string:
            *Pointer + 1
            Integer = 2
          Else
            *Pointer - 1
            Integer = 0
          EndIf
        Else
          Integer = 1
        EndIf
        
        *Start = *Pointer - *Line - 1
        
        If Integer
          Repeat
            Select *Pointer\C
              Case '"'
                If Integer = 2
                  *Pointer + 1
                  If *Pointer\C = '"'
                    *Pointer + 1
                    If *Pointer\C = '"'
                      Break
                    Else
                      String + #Q + #Q
                    EndIf
                  Else
                    String + #Q
                  EndIf
                  *Pointer - 1
                Else
                  Break
                EndIf
              Case '#'
                ; Check symbol:
                *Pointer + 1
                Select *Pointer\C
                  Case 'n'
                    String + #N
                  Case 'q'
                    String + #Q
                  Case 't'
                    String + #T
                  Case 'c'
                    String + #C
                  Case 'f'
                    String + Chr(131)
                  Case 'a'
                    String + Chr(216)
                  Case 'd'
                    String + Chr(149)
                  Case '\'
                    String + GetCurrentDirectory()
                  Case '0' To '2'
                    String + Chr(Val(PeekS(*Pointer, 3)))
                    *Pointer + 2
                  Case '('
                    ; Includes an expression into a string:
                    ; Example: "a = #(1 + 1) = b" -> ("a = " << (1 + 1) << " = b")
                    If Not *OtherTokenQueue
                      *OtherTokenQueue = CreateQueue(@FreeToken())
                      AddBlockToken(*TokenQueue, #Token_Block, #Bracket_Round, *OtherTokenQueue)
                    Else
                      AddIntegerToken(*OtherTokenQueue, #Token_Operator, #Operator_Format)
                    EndIf
                    *Pointer + 1
                    
                    If String
                      AddValueToken(*OtherTokenQueue, #Token_Value, CreateSValue(String))
                      AddIntegerToken(*OtherTokenQueue, #Token_Operator, #Operator_Format)
                    EndIf
                    
                    String = Parser_GetCode(@*Pointer) : _EB(2)
                    Parse(*OtherTokenQueue, @String)   : _EB(2)
                    FreeLastQueueItem(*OtherTokenQueue)
                    
                    *Pointer + 1
                    String = ""
                  Case ':'
                    *Pointer + 1
                    *Start = *Pointer
                    Repeat
                      Select *Pointer\C
                        Case '0' To '9', '-'
                          *Pointer + 1
                        Default
                          Break
                      EndSelect
                    ForEver
                    *Item = Val(PeekS(*Start, *Pointer - *Start))
                    
                    If *Pointer\C = '('
                      If Not *OtherTokenQueue
                        *OtherTokenQueue = CreateQueue(@FreeToken())
                        AddBlockToken(*TokenQueue, #Token_Block, #Bracket_Round, *OtherTokenQueue)
                      Else
                        AddIntegerToken(*OtherTokenQueue, #Token_Operator, #Operator_Format)
                      EndIf
                      *Pointer + 1
                      
                      If String
                        AddValueToken(*OtherTokenQueue, #Token_Value, CreateSValue(String))
                        AddIntegerToken(*OtherTokenQueue, #Token_Operator, #Operator_Format)
                      EndIf
                      
                      *SubTokenQueue = CreateQueue(@FreeToken())
                      AddBlockToken(*OtherTokenQueue, #Token_Block, #Bracket_Round, *SubTokenQueue)
                      
                      String = Parser_GetCode(@*Pointer) : _EB(2)
                      Parse(*SubTokenQueue, @String)     : _EB(2)
                      
                      AddIntegerToken(*OtherTokenQueue, #Token_Control, #Control_Class)
                      AddStringToken(*OtherTokenQueue, #Token_Name, "_form")
                      AddIntegerToken(*OtherTokenQueue, #Token_Control, #Control_Class)
                      AddStringToken(*OtherTokenQueue, #Token_Name, "fix")
                      
                      *SubTokenQueue = CreateQueue(@FreeToken())
                      AddBlockToken(*OtherTokenQueue, #Token_Block, #Bracket_Round, *SubTokenQueue)
                      AddValueToken(*SubTokenQueue, #Token_Value, CreateIValue(*Item))
                      AddIntegerToken(*SubTokenQueue, #Token_Seperator, #Seperator_Block)
                      
                      *Pointer + 1
                      String = ""
                    Else
                      _ErrorB("No bracket used after " + #Q + "#:" + Str(*Item) + #Q)
                    EndIf
                  Case 13
                    ; Ignore newline:
                    *Pointer + 2 + *Start
                  Case '#'
                    String + "#"
                  Default
                    _ErrorB("Invalid character used after " + #Q + "#" + #Q + " in a string: " + #Q + Chr(*Pointer\C) + #Q + ", ASCII = " + #Q + Str(*Pointer\C) + #Q, 2)
                EndSelect
              Case 13
                ; Newline:
                *Pointer + 2 + *Start
                String + #N
              Case 0
                _ErrorB("A string must closed by a second double quote: " + #Q + String + "..." + #Q, 2)
              Default
                ; Put a character in the current string:
                String + Chr(*Pointer\C)
            EndSelect
            *Pointer + 1
          ForEver
        EndIf
        If *OtherTokenQueue
          If String
            AddIntegerToken(*OtherTokenQueue, #Token_Operator, #Operator_Format)
            AddValueToken(*OtherTokenQueue, #Token_Value, CreateSValue(String))
          EndIf
          AddIntegerToken(*OtherTokenQueue, #Token_Seperator, #Seperator_Block)
          *OtherTokenQueue = 0
        Else
          AddValueToken(*TokenQueue, #Token_Value, CreateSValue(String))
        EndIf
      Case '+', '-', '*', '/', '=', '<', '>', '&', '%', '|'
        ; Operator:
        *Start = *Pointer
        *Pointer + 1
        Repeat
          Select *Pointer\C
            Case '+', '-', '*', '/', '=', '<', '>', '&', '%', '|', '!'
              *Pointer + 1
            Default
              *Pointer - 1
              Break
          EndSelect
        ForEver
        ; Read operator:
        String = PeekS(*Start, *Pointer - *Start + 1)
        Select String
          Case "&&"
            AddIntegerToken(*TokenQueue, #Token_Keyword, #Keyword_And)
          Case "||"
            AddIntegerToken(*TokenQueue, #Token_Keyword, #Keyword_Or)
          Case "&"
            AddIntegerToken(*TokenQueue, #Token_Control, #Control_Type)
            AddStringToken(*TokenQueue, #Token_Name, "r")
          Default
            ; Check operator:
            Integer = OperatorID(String)
            If Integer <> -1
              AddIntegerToken(*TokenQueue, #Token_Operator, Integer)
              If OperatorLeftValue(Integer)
                *Pointer = Parser_IgnoreNewLine(*Pointer, @*Line)
              EndIf
            Else
              _ErrorB("Invalid operator: " + #Q + String + #Q)
            EndIf
        EndSelect
      Case '(', '[', '{'
        ; Open a new block:
        AddBlockToken(*TokenQueue, #Token_Block, BracketID(Chr(*Pointer\C)), 0)
        AppendQueueItem(*TokenQueueStack, *TokenQueue)
        *TokenQueue = CreateQueue(@FreeToken())
        *Pointer = Parser_IgnoreNewLine(*Pointer, @*Line)
      Case ')', ']', '}'
        ; Close the current block:
        If *TokenQueueStack\Count
          With *TokenQueue
            If \Last And \Last\Token\Type = #Token_Seperator And \Last\Token\ID = #Seperator_Command
              \Last\Token\ID = #Seperator_Block
            Else
              AddIntegerToken(*TokenQueue, #Token_Seperator, #Seperator_Block)
            EndIf
            Integer = *TokenQueue
            *TokenQueue = PopQueueItem(*TokenQueueStack)
            \Last\Token\Block\TokenQueue = Integer
            If \Last\Token\Block\Bracket <> BracketID(Chr(*Pointer\C))
              _ErrorB("Two different bracktes used for the same block: " + #Q + Bracket(\Last\Token\Block\Bracket) + "..." + Chr(*Pointer\C) + #Q)
            EndIf
          EndWith
        Else
          _ErrorB("Closing bracket without an opened block: " + #Q + "..." + Chr(*Pointer\C) + #Q)
        EndIf
      Case '.'
        ; Single dot or dirpple (ETC) dot:
        *Pointer + 1
        If *Pointer\C = '.' ;= ".."
          *Pointer + 1
          If *Pointer\C = '.' ;= "..."
            AddIntegerToken(*TokenQueue, #Token_Control, #Control_ECT)
          Else
            _ErrorB("Two dots together (" + #Q + ".." + #Q + ") are not allowed.")
          EndIf
        Else
          *Pointer - 1
          AddIntegerToken(*TokenQueue, #Token_Control, #Control_Type)
          *Pointer = Parser_IgnoreNewLine(*Pointer, @*Line)
        EndIf
      Case '\'
        ; Slash:
        *Pointer + 1
        If *Pointer\C = '\' ;= "\\"
          AddIntegerToken(*TokenQueue, #Token_Control, #Control_Previous)
        Else
          *Pointer - 1
          AddIntegerToken(*TokenQueue, #Token_Control, #Control_Next)
        EndIf
      Case ':'
        ; Double dot:
        *Pointer + 1
        If *Pointer\C = ':' ;= "::"
          AddIntegerToken(*TokenQueue, #Token_Control, #Control_ObjectClass)
        Else
          *Pointer - 1
          AddIntegerToken(*TokenQueue, #Token_Control, #Control_Class)
        EndIf
      Case '@'
        ; Address:
        AddIntegerToken(*TokenQueue, #Token_Control, #Control_Address)
      Case ';'
        ; End of command:
        With *TokenQueue
          If \Last And Not(\Last\Token\Type = #Token_Seperator And \Last\Token\Type = #Seperator_Command)
            AddIntegerToken(*TokenQueue, #Token_Seperator, #Seperator_Command)
          EndIf
        EndWith
        If SingleLine
          ProcedureReturn *Pointer
        EndIf
      Case ','
        ; Argument seperator:
        AddIntegerToken(*TokenQueue, #Token_Seperator, #Seperator_Argument)
        *Pointer = Parser_IgnoreNewLine(*Pointer, @*Line)
      Case '#'
        ; Preprocessor functions:
        *Pointer + 1
        Select *Pointer\C
          Case 'a' To 'z', '_', 'A' To 'Z', 192 To 255
            ; Symbol:
            *Start = *Pointer
            *Pointer = Parser_GetName(*Pointer)
            ; Read symbol:
            String = PeekS(*Start, *Pointer - *Start + 1)
            ; Evaluate symbol:
            *Pointer = ParseSymbol(*TokenQueue, String, *Pointer)
            _EB(1, "For preprocessor symbol: " + #Q + String + #Q)
          Case '('
            ; Evaluate an expression:
            *Pointer + 1
            String = Parser_GetCode(@*Pointer)             : _EB()
            *Pointer + 1
            Integer = Execute(String, *PreprocessorObject) : _EB()
            AddValueToken(*TokenQueue, #Token_Value, DuplicateValue(Integer))
          Case '{'
            ; Open a level in the preprocessor object:
            *Pointer + 1
            *Start = *Pointer
            Parser_ScanBlock(@*Pointer)                                            : _EB()
            ExecuteCommands(PeekS(*Start, *Pointer - *Start), *PreprocessorObject) : _EB()
            *Pointer + 1 ;Ignore the "}"
          Case '$'
            ; Add an ascii code:
            *Pointer + 1
            AddValueToken(*TokenQueue, #Token_Value, CreateIValue(*Pointer\C))
          Case 13
            ; Ignore end of line:
            *Pointer + 1
            *Line = *Pointer + 1
          Default
            _ErrorB("Invalid character used after " + #Q + "#" + #Q + ":" + #N + "- char:  " + Chr(*Pointer\C) + #N + "- ASCII: " + Str(*Pointer\C))
        EndSelect
      Case 39
        ; Comment:
        *Pointer = Parser_IgnoreComment(*Pointer) : _E("", Break)
      Case 13
        ; End of line:
        With *TokenQueue
          If \Last And Not(\Last\Token\Type = #Token_Seperator And \Last\Token\Seperator = #Seperator_Command)
            AddIntegerToken(*TokenQueue, #Token_Seperator, #Seperator_Command)
          EndIf
        EndWith
        *Pointer + 1 ;Ignore ASCII-10!
        If SingleLine
          ProcedureReturn *Pointer
        EndIf
        *Line = *Pointer + 1
      Case 10
        ; End of line without ASCII-13:
        With *TokenQueue
          If \Last And Not(\Last\Token\Type = #Token_Seperator And \Last\Token\Seperator = #Seperator_Command)
            AddIntegerToken(*TokenQueue, #Token_Seperator, #Seperator_Command)
          EndIf
        EndWith
        If SingleLine
          ProcedureReturn *Pointer
        EndIf
        *Line = *Pointer + 1
      Case ' ', 9
        ; Ignore characters:
        *Pointer + 1
        Continue
      Case '^'
        ; Power:
        AddIntegerToken(*TokenQueue, #Token_Operator, #Operator_Power)
        *Pointer = Parser_IgnoreNewLine(*Pointer, @*Line)
      Case '', ''
        ; Exponentiate to 2 or 3:
        AddIntegerToken(*TokenQueue, #Token_Operator, #Operator_Power)
        AddValueToken(*TokenQueue, #Token_Value, CreateIValue(*Pointer\C - 176)) ;ASCII 178 = ""
      Case ''
        ; Convertion of degrees to radians:
        AddIntegerToken(*TokenQueue, #Token_Operator, #Operator_Multiply)
        AddValueToken(*TokenQueue, #Token_Value, CreateFValue(#PI/180))
      Case 0
        ; End of file:
        With *TokenQueue
          If Not(\Last And \Last\Token\Type = #Token_Seperator And \Last\Token\Type = #Seperator_Command)
            AddIntegerToken(*TokenQueue, #Token_Seperator, #Seperator_Command)
          EndIf
        EndWith
        Break
      Case 1
        ; Self forzed end of code:
        Break
      Default
        ; Invalid character:
        _ErrorB("Invalid character used:" + #N + "- char:  " + Chr(*Pointer\C) + #N + "- ASCII: " + Str(*Pointer\C))
    EndSelect
    *Pointer + 1
  ForEver
  
  If *TokenQueueStack\Count And Not ErrorMessage
    _Error(Form(*TokenQueueStack\Count, "bracket is", "brackets are") + " Not closed")
  EndIf
  FreeQueue(*TokenQueueStack)
  
  ProcedureReturn *Pointer
EndProcedure
Procedure.i ParseFile(Name.s, *TokenQueue.Queue)
  Protected File.i, *Code
  
  ; Add the extension ".op_c", if not aviable:
  If Not GetExtensionPart(Name)
    Name + ".op_c"
  EndIf
  
  ; Find file by two tests:
  File = ReadFile(#PB_Any, Name)
  If Not File
    File = ReadFile(#PB_Any, GetCurrentDirectory() + "Includes\" + Name)
  EndIf
  
  If File
    *Code = AllocateMemory(Lof(File) + 1)
    ReadData(File, *Code, Lof(File))
    
    Parse(*TokenQueue, *Code)
    FreeMemory(*Code)
    
    CloseFile(File)
    _E("in source file: " + #Q + Name + #Q, FreeQueue(*TokenQueue))
    ProcedureReturn #True
  Else
    _Error("File not found: " + #Q + Name + #Q)
  EndIf
EndProcedure
Procedure.i ParseSymbol(*TokenQueue.Queue, Name.s, *Pointer.Character)
  Protected *Object.Object
  Protected String.s, Integer.i
  Protected *ArgumentQueue.Queue, *Item.QueueItem, *Value.Value
  
  ; Check predefined symbols:
  Select CamelCase(Name)
    Case "del"
      *ArgumentQueue = Parser_GetArguments(@*Pointer) : _E()
      *Item = *ArgumentQueue\First
      While *Item
        Integer = ExecuteObject(*Item\Value\String\S, *PreprocessorObject, 0, 0, #True) : _E()
        FreeObject(Integer)                                                       : _E()
        _Move(*Item, N)
      Wend
    Case "if"
      *Pointer = ParseIfControl(*TokenQueue, *Pointer) : _E()
    Case "include", "inc"
      *ArgumentQueue = Parser_GetArguments(@*Pointer) : _E()
      *Item = *ArgumentQueue\First
      While *Item
        String = ExecuteS(*Item\Value\String\S, 0, #True) : _E()
        ParseFile(String, *TokenQueue)                    : _E()
        _Move(*Item, N)
      Wend
      _E()
    Case "date"
      AddValueToken(*TokenQueue, #Token_Value, CreateSValue(FormatDate("%dd.%mm.%yyyy", Date())))
    Case "time"
      AddValueToken(*TokenQueue, #Token_Value, CreateSValue(FormatDate("%ss:%ii:%hh", Date())))
    Case "out"
      *ArgumentQueue = Parser_GetArguments(@*Pointer) : _E()
      *Item = *ArgumentQueue\First
      While *Item
        *Value = Execute(*Item\Value\String\S, *PreprocessorObject, #True) : _E()
        String = FormatValue(*Value)                                       : _E()
        OutputFL(#Symbol_Output, String)
        _Move(*Item, N)
      Wend
    Case "selectFile"
      *ArgumentQueue = Parser_GetArguments(@*Pointer, #False) : _E()
      If *ArgumentQueue And *ArgumentQueue\Count = 1
        String = ValueS(*ArgumentQueue\First\Value) : _E()
      Else
        String = "Select file"
      EndIf
      AddValueToken(*TokenQueue, #Token_Value, CreateSValue(OpenFileRequester(String, #OpernoPath, "*.*", 0)))
    Case "consoleApp"
      Parse(*TokenQueue, @"stream{openConsole;outputSymbols(0)}")
      
    Default
      ; Check other symbols:
      *Object = ExecuteObject(Name, *PreprocessorObject, 0, 0, #False) : _E()
      If Not *Object
        *Pointer + 1
        Name + Parser_GetCode(@*Pointer) : _E()
        
        *Object = Abs(ExecuteObject(Name, *PreprocessorObject, #RunPath_Define, 0, #True)) : _E()
        
        If *Object\Type = #Object_Variable And *Object\Variable\Value\Class = *VariantClass
          Name = ObjectPath(*Object)
          FreeObject(*Object)
          _Error("Symbol not definied.")
        EndIf
        
        ProcedureReturn *Pointer
      EndIf
      ; Evaluate symbol:
      Select *Object\Type
        Case #Object_Variable
          ; Constant value:
          AddValueToken(*TokenQueue, #Token_Value, DuplicateValue(*Object\Variable\Value))
        Case #Object_Function
          ; Macro:
          *ArgumentQueue = Parser_GetArguments(@*Pointer, #False)
          ; Call macro:
          *Value = Call(*Object, *ArgumentQueue) : _E()
          ; Parse the returned string:
          If *Value\Class = *StringClass
            Parse(*TokenQueue, @*Value\String\S)
            FreeLastQueueItem(*TokenQueue) ;Remove the invalid Seperator
          Else
            _Error("Result of the macro is not a string!")
          EndIf
          
          *ArgumentQueue = #Null
        Case #Object_Group
          _Error("Groups are not allowed as symbol.")
      EndSelect
  EndSelect
  
  If *ArgumentQueue > 0
    FreeQueue(*ArgumentQueue)
  EndIf
  
  ProcedureReturn *Pointer
EndProcedure
Procedure.i ParseIfControl(*TokenQueue.Queue, *Pointer.Character)
  Protected *ControlTokenQueue.Queue, *Token.Token
  Protected Code.s, *SelectedToken.Token, *SelectedTokenQueue.Queue
  
  *Pointer + 1
  Code = Parser_GetCode(@*Pointer) : _E()
  
  *ControlTokenQueue = CreateQueue(@FreeToken())
  Parse(*ControlTokenQueue, @Code) : _E("", FreeQueue(*ControlTokenQueue))
  
  OpenLevel(*ControlTokenQueue, 0, *PreprocessorObject, 0)
    If IsValueTrue(RunControlBlock())
     *SelectedTokenQueue = *Current\TokenItem\Token\Block\TokenQueue
    Else
      _N()
      Repeat
        _SN(*Token)
        If *Token\Type = #Token_Keyword
          Select *Token\Keyword
            Case #Keyword_ElseIf
              If IsValueTrue(RunControlBlock())
                *SelectedToken = *Current\TokenItem\Token
                Break
              Else
                _N() ;Ignore controll-block
              EndIf
              Continue
            Case #Keyword_Else
              *SelectedToken = *Current\TokenItem\Token
          EndSelect
        EndIf
        Break
      ForEver
    EndIf
  CloseLevel()
  
  If *SelectedToken
    If *SelectedToken\Type = #Token_Block
      CopyQueue(*SelectedToken\Block\TokenQueue, *TokenQueue, @CopyToken())
      FreeLastQueueItem(*TokenQueue)
    Else
      _Error("A block is missing in a preprocessor if-controll.", FreeQueue(*ControlTokenQueue))
    EndIf
  EndIf
  FreeQueue(*ControlTokenQueue)
  
  ProcedureReturn *Pointer
EndProcedure

Procedure.i Parser_IgnoreNewLine(*Pointer.Character, *Line)
  *Pointer + 1
  Repeat
    Select *Pointer\C
      Case 13
        ; Ignore end of line:
        If *Line
          PokeI(*Line, *Pointer + 2)
        EndIf
        ProcedureReturn *Pointer + 1
      Case 10
        If *Line
          PokeI(*Line, *Pointer + 1)
        EndIf
        ProcedureReturn *Pointer
      Case ' ', 9
        ; Ignore characters:
        *Pointer + 1
        Continue
      Case 39
        ; Jump over the whole comment:
        *Pointer = Parser_IgnoreComment(*Pointer)
      Default
        ProcedureReturn *Pointer - 1
    EndSelect
    *Pointer + 1
  ForEver
EndProcedure
Procedure.i Parser_IgnoreComment(*Pointer.Character)
  Protected Level.i
  
  *Pointer + 1
  Select *Pointer\C
    Case '{'
      ; Comment block:
      *Pointer + 1
      Repeat
        Select *Pointer\C
          Case 39
            *Pointer + 1
            If *Pointer\C = '{'
              ; Open a level:
              Level + 1
            Else
              *Pointer - 1
            EndIf
          Case '}'
            *Pointer + 1
            If *Pointer\C = 39
              ; Close the current level:
              If Level = 0
                ; End of the comment block:
                ProcedureReturn *Pointer
              Else
                Level - 1
              EndIf
            Else
              *Pointer - 1
            EndIf
          Case 0
            _Error("A comment block is not closed.")
        EndSelect
        *Pointer + 1
      ForEver
    Default
      ; Single line comment:
      Repeat
        Select *Pointer\C
          Case 13, 0, 1, 10
            ; End of the comment:
            ProcedureReturn *Pointer - 1
        EndSelect
        *Pointer + 1
      ForEver
  EndSelect
EndProcedure
Procedure.i Parser_IgnoreString(*Pointer.Character)
  Protected TopLevel.i
  
  *Pointer + 1
  If *Pointer\C = '"'
    *Pointer + 1
    If *Pointer\C = '"'
      *Pointer + 1
      TopLevel = #True
    Else
      ProcedureReturn *Pointer - 1
    EndIf
  EndIf
  
  Repeat
    Select *Pointer\C
      Case '"'
        If TopLevel
          *Pointer + 1
          If *Pointer\C = '"'
            *Pointer + 1
            If *Pointer\C = '"'
              ProcedureReturn *Pointer
            EndIf
          EndIf
        Else
          ProcedureReturn *Pointer
        EndIf
      Case '#'
        *Pointer + 1
        If *Pointer\C = '('
          *Pointer + 1
          Parser_GetCode(@*Pointer)
          *Pointer + 1
        EndIf
      Case 0
        _Error("A string must closed by a second double quote.")
    EndSelect
    *Pointer + 1
  ForEver
EndProcedure
Procedure.i Parser_CountLevel(Code.s)
  Protected *Pointer.Character, Level.i
  
  *Pointer = @Code
  While *Pointer\C
    Select *Pointer\C
      Case '"'
        *Pointer = Parser_IgnoreString(*Pointer) : _E()
      Case 39
        *Pointer = Parser_IgnoreComment(*Pointer) : _E()
      Case '(', '[', '{'
        Level + 1
      Case ')', ']', '}'
        Level - 1
    EndSelect
    *Pointer + 1
  Wend
  ProcedureReturn Level
EndProcedure
Procedure.s Parser_GetCode(*Pointer.All, EnableArguments.i = #False)
  Protected *Start.Character, Level.i
  
  *Start = *Pointer\P
  Repeat
    Select *Pointer\P\C
      Case '(', '[', '{'
        Level + 1
      Case ')', ']', '}'
        If Level > EnableArguments
          Level - 1
        ElseIf EnableArguments
          Break
        Else
          *Pointer\P - 1
          Break
        EndIf
      Case '"'
        *Pointer\P = Parser_IgnoreString(*Pointer\P) : _E()
      Case ',', ';', 13, 10
        If Level = 0
          *Pointer\P - 1
          Break
        EndIf
      Case 0, 1
        *Pointer\P - 1
        Break
      Case 39
        *Pointer\P = Parser_IgnoreComment(*Pointer\P) : _E()
    EndSelect
    *Pointer\P + 1
  ForEver
  
  ProcedureReturn PeekS(*Start, *Pointer\P - *Start + 1)
EndProcedure
Procedure.i Parser_ScanBlock(*Pointer.All)
  Protected *Start.Character, Level.i
  
  *Start = *Pointer\P
  Repeat
    Select *Pointer\P\C
      Case '(', '[', '{'
        Level + 1
      Case ')', ']', '}'
        If Level = 0
          Break
        Else
          Level -1
        EndIf
      Case '"'
        *Pointer\P = Parser_IgnoreString(*Pointer\P) : _E()
      Case 39
        *Pointer\P = Parser_IgnoreComment(*Pointer\P) : _E()
      Case 0, 1
        *Pointer\P - 1
        Break
    EndSelect
    *Pointer\P + 1
  ForEver
EndProcedure
Procedure.i Parser_GetArguments(*Pointer.All, EnableError.i = #True)
  Protected *ArgumentQueue.Queue, Arguments.s, *ArgumentPointer.Character
  Protected Code.s
  
  *Pointer\P + 1
  Repeat
    Select *Pointer\P\C
      Case '('
        Arguments = Parser_GetCode(*Pointer, #True)
        If RemoveString(Arguments, " ") <> "()"
          *ArgumentQueue = CreateQueue(#Queue_SaveMemory)
          *ArgumentPointer = @Arguments
          Repeat
            *ArgumentPointer + 1
            Code = Parser_GetCode(@*ArgumentPointer) : _E("", FreeQueue(*ArgumentQueue))
            *ArgumentPointer + 1
            AppendQueueItem(*ArgumentQueue, PushSValue(Trim(Code)))
          Until *ArgumentPointer\C <> ','
          ProcedureReturn *ArgumentQueue
        Else
          ProcedureReturn -1
        EndIf
      Case ' ', 9
        ;Do nothing
      Default
        If EnableError
          _Error("No argument queue found.")
        Else
          *Pointer\P - 1
          ProcedureReturn #False
        EndIf
    EndSelect
    *Pointer\P + 1
  ForEver
EndProcedure
Procedure.i Parser_GetName(*Pointer.Character)
  Protected Level.i, *Start.Character, *WordStart.Character
  
  Repeat
    Select *Pointer\C
      Case 'a' To 'z', 'A' To 'Z', '_', '0' To '9', 192 To 255, '!', '$', '?'
        ; Valid...
      Case ' '
        If *WordStart
          If KeywordID(PeekS(*WordStart, *Pointer - *WordStart)) <> #None
            *Pointer = *WordStart - 1
            Break
          EndIf
        EndIf
        ; Check space:
        If Level = 0
          *Start = *Pointer - 1
          *Pointer + 1
          ; Ignore all spaces:
          Repeat
            Select *Pointer\C
              Case ' '
                *Pointer + 1
              Case 0
                ProcedureReturn *Start
              Default
                Break
            EndSelect
          ForEver
          ; Check next character:
          Select *Pointer\C
            Case 'a' To 'z', 'A' To 'Z', '_'
              *WordStart = *Pointer
              ; Valid...
            Default
              ProcedureReturn *Start
          EndSelect
        EndIf
      Case '\'
        ; Next child object:
        *Pointer + 1
        If *Pointer\C = '\' ;= "\\"
          *Pointer + 1
        EndIf
        If Level = 0
          If *Pointer\C = ' '
            *Pointer + 1
            Repeat
              Select *Pointer\C
                Case ' '
                  *Pointer + 1
                Case 0
                  *Pointer - 1
                  Break 2
                Default
                  Break
              EndSelect
            ForEver
          EndIf
          Select *Pointer\C
            Case 'a' To 'z', 'A' To 'Z', '_'
              ; Valid...
            Case '('
              Level + 1
            Default
              *Pointer - 1
              Break
          EndSelect
        Else
          *Pointer - 1
        EndIf
      Case '"'
        If Level
          *Pointer = Parser_IgnoreString(*Pointer) : _E()
        Else
          *Pointer - 1
          Break
        EndIf
      Case '(', '[', '{'
        ; Open a level or break:
        If Level
          Level + 1
        Else
          *Pointer - 1
          Break
        EndIf
      Case ')', ']', '}'
        ; Close a level or break:
        If Level
          Level - 1
        Else
          *Pointer - 1
          Break
        EndIf
      Default
        ; End of the name:
        If Level = 0
          If *WordStart
            If KeywordID(PeekS(*WordStart, *Pointer - *WordStart)) <> #None
              *Pointer = *WordStart - 1
              Break
            EndIf
          EndIf
          *Pointer - 1
          Break
        EndIf
    EndSelect
    *Pointer + 1
  ForEver
  
  ProcedureReturn *Pointer
EndProcedure



;******************************************
; Preprozessor
;******************************************
; Hier sind alle ntigen Prprozessor-Objekte eingebuden:

Procedure.i Pre_HorinzontalLine()
  ProcedureReturn PushSValue("out " + #Q + "------------------------------" + #Q)
EndProcedure
Procedure.i Pre_Opposite()
  Protected Variable.String, Result.s
  
  While PopSValue(@Variable)
    If Result
      Result + "; "
    EndIf
    Result + Variable\S + " = -" + Variable\S
  Wend
  ProcedureReturn PushSValue(Result)
EndProcedure
Procedure.i Pre_Operator()
  Protected Operator.String, Result.s, FriendMethod.i
  
  If PopSValue(@Operator)
    If Mid(Operator\S, 1, 1) = "f"
      FriendMethod = #True
      Operator\S = Mid(Operator\S, 2)
    EndIf
    Select Operator\S
      Case "+"
        Result = "add"
      Case "-"
        Result = "sub"
      Case "*"
        Result = "mul"
      Case "/"
        Result = "div"
      Case "**"
        Result = "pow"
      Case "+="
        Result = "add!"
      Case "-="
        Result = "sub!"
      Case "*="
        Result = "mul!"
      Case "/="
        Result = "div!"
    EndSelect
    If FriendMethod
      ProcedureReturn PushSValue("_f" + UCase(Left(Result, 1)) + Mid(Result, 2))
    Else
      ProcedureReturn PushSValue("_" + Result)
    EndIf
  EndIf
EndProcedure

DataSection
  PreprocessorObject:
    Data.c #Object_Variable : Data.s "true" ;{
      Data.c #Type_Bool
      Data.b #True
    ;}
    Data.c #Object_Variable : Data.s "false" ;{
      Data.c #Type_Bool
      Data.b #False
    ;}
    Data.c #Object_Variable : Data.s "pi" ;{
      Data.c #Type_Double
      Data.d 3.14159265359
    ;}
    Data.c #Object_Variable : Data.s "e" ;{
      Data.c #Type_Double
      Data.d 2.71828182846
    ;}
    
    Data.c #Object_Group : Data.s "token" ;{
      Data.c #Object_Variable : Data.s "indentifer" ;{
        Data.c #Type_Integer
        Data.i #Token_Name
      ;}
      Data.c #Object_Variable : Data.s "value" ;{
        Data.c #Type_Integer
        Data.i #Token_Value
      ;}
      Data.c #Object_Variable : Data.s "operator" ;{
        Data.c #Type_Integer
        Data.i #Token_Operator
      ;}
      Data.c #Object_Variable : Data.s "block" ;{
        Data.c #Type_Integer
        Data.i #Token_Block
      ;}
      Data.c #Object_Variable : Data.s "seperator" ;{
        Data.c #Type_Integer
        Data.i #Token_Seperator
      ;}
      Data.c #Object_Variable : Data.s "control" ;{
        Data.c #Type_Integer
        Data.i #Token_Control
      ;}
      Data.c #Object_Variable : Data.s "keyword" ;{
        Data.c #Type_Integer
        Data.i #Token_Keyword
      ;}
      Data.c #Null
    ;}
    
    Data.c #Object_Function : Data.s "opo" ;{
      Data.i @Pre_Opposite()
    ;}
    Data.c #Object_Function : Data.s "hl" ;{
      Data.i @Pre_HorinzontalLine()
    ;}
    Data.c #Object_Function : Data.s "operator" ;{
      Data.i @Pre_Operator()
    Data.s "op"
    ;}
      
    Data.c #Object_Group : Data.s "symbol" ;{
      Data.c #Object_Variable : Data.s "none" ;{
        Data.c #Type_Integer
        Data.i #Symbol_None
      ;}
      Data.c #Object_Variable : Data.s "default" ;{
        Data.c #Type_Integer
        Data.i #Symbol_Default
      ;}
      Data.c #Object_Variable : Data.s "etc" ;{
        Data.c #Type_Integer
        Data.i #Symbol_ETC
      ;}
      Data.c #Object_Variable : Data.s "point" ;{
        Data.c #Type_Integer
        Data.i #Symbol_Point
      ;}
      Data.c #Object_Variable : Data.s "output" ;{
        Data.c #Type_Integer
        Data.i #Symbol_Output
      ;}
      Data.c #Object_Variable : Data.s "input" ;{
        Data.c #Type_Integer
        Data.i #Symbol_Input
      ;}
      Data.c #Object_Variable : Data.s "error" ;{
        Data.c #Type_Integer
        Data.i #Symbol_Error
      ;}
      Data.c #Object_Variable : Data.s "message" ;{
        Data.c #Type_Integer
        Data.i #Symbol_Message
      ;}
      Data.c #Null
    ;}
    
    Data.c #Object_Variable : Data.s "version" ;{
      Data.c #Type_String
      Data.s #Version
    ;}
    Data.c #Null
EndDataSection
; IDE Options = PureBasic 4.41 (Windows - x86)
; CursorPosition = 1301
; FirstLine = 807
; Folding = HAACEAw
; EnableXP
; CompileSourceDirectory